'''pastry_client handles the interaction with the chef server api'''

import os
import yaml
import requests
from contextlib import contextmanager

from pastry.utils.auth import signed_headers
from pastry.exceptions import HttpError


HTTP_METHODS = {
    'GET': requests.get,
    'POST': requests.post,
    'PUT': requests.put,
    'DELETE': requests.delete
}


class PastryClient(object):
    '''
    PastryClient is used by the resources to send requests to chef.
    You can use the PastryClient to initialize the config to use for the
    chef server.
    '''
    server = None
    organization = None
    client = None
    keypath = None
    verify = None
    initialized = False

    @classmethod
    def initialize(cls, server, organization, client, keypath, verify):
        '''
        Initializes the server connection info

        :param server: The url for the chef server e.g. https://my.chef.server
        :param organization: The name of the cheff org to use
        :param client: The client/username to use
        :param keypath: The path to the pem for the client/user
        :param verify: Verify the ssl cert on requests
        :type server: string
        :type organization: string
        :type client: string
        :type keypath: string
        :type verify: boolean
        '''
        cls.server = server
        cls.organization = organization
        cls.client = client
        cls.keypath = keypath
        cls.verify = verify
        cls.initialized = True

    @classmethod
    @contextmanager
    def context(cls, organization=organization):
        '''
        Context manager to override the default org

        :param organization: The name of the org to use for the command
        :type organization: string

        .. code-block::

            with PastryClient.context(organization='myorg'):
                print Nodes.exists('mynode')
        '''
        if not cls.initialized:
            cls.load_config()
        default_org = cls.organization
        cls.organization = organization
        yield
        cls.organization = default_org

    @classmethod
    def load_config(cls, config_path=None):
        '''
        Load server config from a yaml file

        If no config_path is specified it will try to load config from
        ``$HOME/.chef/pastry.yaml``

        .. note::

            This method is automatically called if an http requests is made and
            the client has not yet been initialised

        :param config_path: The path to the config file on the local filesystem
        :type config_path: string
        '''
        if not config_path:
            config_path = os.path.join(
                os.environ['HOME'], '.chef', 'pastry.yaml')
        if not os.path.exists(config_path):
            raise ValueError(
                'PastryClient not initialized and %s does not exist' %
                config_path
            )
        with open(config_path, 'r') as config_file:
            config = yaml.safe_load(config_file.read())
        cls.initialize(
            config['server'],
            config['organization'],
            config['client'],
            config['keypath'],
            config.get('verify', True)
        )

    @classmethod
    def get_url(cls, endpoint):
        '''
        Fetches the server and updates the endpoint with the organization

        :param endpoint: The endpoint being called on the chef server
        :type endpoint: string
        :return: The server url and updated endpoint
        :rtype: tuple
        '''
        if not cls.initialized:
            cls.load_config()
        return cls.server, (endpoint % {'org': cls.organization})

    @classmethod
    def call(cls, endpoint, method='GET', data=None):
        '''
        Send an http request to the chef server api

        :param endpoint: The endpoint for the resource being called
        :param method: The HTTP method
        :param data: The body of the request
        :type endpoint: string
        :type method: string
        :type data: hash
        :return: The json response from the server
        :rtype: hash
        '''
        server, path = cls.get_url(endpoint)
        headers = signed_headers(
            cls.client,
            cls.keypath,
            path.split('?', 1)[0],
            method=method,
            data=data
        )
        kwargs = {
            'headers': headers,
            'verify': cls.verify
        }
        if data:
            kwargs['json'] = data
        resp = HTTP_METHODS[method]('%s%s' % (server, path), **kwargs)
        if not resp.ok:
            raise HttpError(resp.text, resp.status_code)
        return resp.json()

    @classmethod
    def status(cls):
        '''
        Check the chef server status

        :return: The json response form the server
        :type: hash
        '''
        resp = requests.get('%s/_status' % cls.server, verify=cls.verify)
        if not resp.ok:
            raise HttpError(resp.text, resp.status_code)
        return resp.json()



'''pastry_rsa provides a signing method that works with openssl rsautl'''

import rsa


def sign(message, raw_key):
    '''sign a request the same way openssl rsautl would'''
    priv_key = rsa.PrivateKey.load_pkcs1(raw_key)
    keylength = rsa.common.byte_size(priv_key.n)
    padded = rsa.pkcs1._pad_for_signing(message, keylength)
    payload = rsa.transform.bytes2int(padded)
    encrypted = rsa.core.encrypt_int(payload, priv_key.d, priv_key.n)
    return rsa.transform.int2bytes(encrypted, keylength)



'''auth provides helper methods for setting up the headers required by chef'''

import json
from base64 import b64encode
from hashlib import sha1
from datetime import datetime
from .pastry_rsa import sign


CANONICAL = '''Method:%(method)s
Hashed Path:%(hashed_path)s
X-Ops-Content-Hash:%(hashed_body)s
X-Ops-Timestamp:%(timestamp)s
X-Ops-UserId:%(client)s'''


def hashencode(content):
    '''sha1 encode text'''
    hashed = sha1(content.encode()).digest()
    return encode(hashed)


def encode(hashed):
    '''base64 encode bytes and break it up into 60 character chunks'''
    based = b64encode(hashed)
    return [based[i:i + 60].decode() for i in range(0, len(based), 60)]


def authorization_headers(keypath, canonical_source):
    '''create tehe authorization headers'''
    with open(keypath, 'rb') as keyfile:
        signature = sign(canonical_source.encode(), keyfile.read())
    parts = encode(signature)
    return {
        'X-ops-authorization-%s' % (i + 1): val for i, val in enumerate(parts)
    }


def signed_headers(client, keypath, path, method='GET', data=None):
    '''get the signed request headers for chef api'''
    hashed_path = ''.join(hashencode(path))
    body = ''
    if data:
        body = json.dumps(data)
    hashed_body = '\n'.join(hashencode(body))
    timestamp = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')

    canonical_source = CANONICAL % {
        'method': method,
        'hashed_path': hashed_path,
        'hashed_body': hashed_body,
        'timestamp': timestamp,
        'client': client
    }

    headers = {
        'Accept': 'application/json',
        'X-ops-sign': 'version=1.0',
        'X-ops-userid': client,
        'X-ops-timestamp': timestamp,
        'X-ops-content-hash': hashed_body,
        'X-chef-version': '12.8.0'
    }
    headers.update(authorization_headers(keypath, canonical_source))
    return headers



'''exceptions for pastry events'''

class HttpError(Exception):
    '''
    Exception calling http request -adds teh ststus code)
    '''
    def __init__(self, message, statuscode, *args, **kwargs):
        '''
        Create the Exception
        '''
        self.statuscode = statuscode
        super(HttpError, self).__init__(message, *args, **kwargs)

    def __str__(self):
        '''
        Show the exception and statuscode
        '''
        return '(%s) %s' % (self.statuscode, self.message)



import mock
import unittest

from pastry.resources.search import Search


class SearchTestCase(unittest.TestCase):

    @mock.patch('pastry.resources.search.Base.index', return_value='index')
    def test_index(self, base):
        self.assertEqual(Search.index(), 'index')

    @mock.patch('pastry.resources.search.urllib.urlencode', return_value='query_string')
    @mock.patch('pastry.resources.search.PastryClient.call', return_value='results')
    @mock.patch('pastry.resources.search.Base.escape_query', return_value='escaped')
    def test_run(self, escape_query, call, urlencode):
        self.assertEqual(Search.run(
            'index',
            query='query',
            rows=1,
            start=1,
            filters={'name': 'name'}
        ), 'results')
        urlencode.assert_called_with({'query': 'query', 'rows': 1, 'start': 1})
        escape_query.assert_called_with('query_string')
        call.assert_called_with(
            '/organizations/%(org)s/search/index?escaped',
            method='POST',
            data={'name': 'name'}
        )



import mock
import unittest

from pastry.utils.pastry_rsa import sign


class PastryRsaTestCase(unittest.TestCase):

    @mock.patch('pastry.utils.pastry_rsa.rsa')
    def test_sign(self, rsa):
        private = mock.MagicMock()
        private.n = 'n'
        private.d = 'd'
        rsa.PrivateKey.load_pkcs1.return_value = private
        rsa.common.byte_size.return_value = 'byte_size'
        rsa.pkcs1._pad_for_signing.return_value = 'padded'
        rsa.transform.bytes2int.return_value = 'payload'
        rsa.core.encrypt_int.return_value = 'encrypted'
        rsa.transform.int2bytes.return_value = 'int2bytes'
        self.assertEqual('int2bytes', sign('message', 'key'))
        rsa.PrivateKey.load_pkcs1.assert_called_with('key')
        rsa.common.byte_size.assert_called_with('n')
        rsa.pkcs1._pad_for_signing.assert_called_with('message', 'byte_size')
        rsa.transform.bytes2int.assert_called_with('padded')
        rsa.core.encrypt_int.assert_called_with('payload', 'd', 'n')
        rsa.transform.int2bytes.assert_called_with('encrypted', 'byte_size')



import mock
import unittest

from pastry.pastry_client import PastryClient


class PastryClientTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        PastryClient.server = None
        PastryClient.organization = None
        PastryClient.client = None
        PastryClient.keypath = None

    def test_initialize(self):
        PastryClient.initialize('server', 'organization', 'client', 'keypath', 'verify')
        self.assertEqual(PastryClient.server, 'server')
        self.assertEqual(PastryClient.organization, 'organization')
        self.assertEqual(PastryClient.client, 'client')
        self.assertEqual(PastryClient.keypath, 'keypath')
        self.assertEqual(PastryClient.verify, 'verify')

    @mock.patch('pastry.pastry_client.PastryClient.load_config')
    def test_organization(self, load_config):
        current = PastryClient.organization
        PastryClient.initialized = False
        with PastryClient.context(organization='org'):
            self.assertEqual(PastryClient.organization, 'org')
        self.assertEqual(PastryClient.organization, current)
        self.assertEqual(load_config.call_count, 1)

    @mock.patch('pastry.pastry_client.PastryClient.initialize')
    @mock.patch('pastry.pastry_client.os')
    @mock.patch('pastry.pastry_client.yaml')
    def test_load_config(self, yaml, os, initialize):
        os.path.join.return_value = 'config_path'
        os.path.exists.return_value = True
        yaml.safe_load.return_value = {
            'server': 'server',
            'organization': 'organization',
            'client': 'client',
            'keypath': 'keypath',
            'verify': 'verify'
        }
        with mock.patch('__builtin__.open', mock.mock_open(read_data='config')):
            PastryClient.load_config()
            initialize.assert_called_with(
                'server', 'organization', 'client', 'keypath', 'verify')
        os.path.exists.return_value = False
        self.assertRaises(ValueError, PastryClient.load_config)

    @mock.patch('pastry.pastry_client.PastryClient.load_config')
    def test_get_url(self, load_config):
        self.assertEqual((None, 'endpoint'), PastryClient.get_url('endpoint'))

    @mock.patch('pastry.pastry_client.HTTP_METHODS')
    @mock.patch('pastry.pastry_client.signed_headers')
    @mock.patch('pastry.pastry_client.PastryClient.get_url')
    def test_call(self, get_url, signed_headers, methods):
        get_url.return_value = ('server', 'path')
        signed_headers.return_value = 'headers'
        response = mock.MagicMock()
        response.ok = True
        methods['GET'].return_value = response
        PastryClient.call('endpoint')
        methods['GET'].assert_called_with('serverpath', headers='headers', verify=PastryClient.verify)
        PastryClient.call('endpoint', method='POST', data={'key': 'value'})
        methods['POST'].assert_called_with(
            'serverpath',
            headers='headers',
            json={'key': 'value'},
            verify=PastryClient.verify
        )
        response.ok = False
        self.assertRaises(Exception, PastryClient.call, 'endpoint')

    @mock.patch('pastry.pastry_client.requests.get')
    def test_status(self, get):
        resp = mock.MagicMock()
        resp.ok = True
        resp.json.return_value = {'json': 'content'}
        get.return_value = resp
        self.assertEqual(PastryClient.status(), resp.json())
        get.assert_called_with('%s/_status' % PastryClient.server, verify=PastryClient.verify)
        resp.ok = False
        self.assertRaises(Exception, PastryClient.status)



import mock
import unittest

from pastry.resources.groups import Groups


class GroupsTestCase(unittest.TestCase):

    @mock.patch('pastry.resources.groups.Base.index', return_value='index')
    def test_index(self, base):
        self.assertEqual(Groups.index(), 'index')

    @mock.patch('pastry.resources.groups.Base.get', return_value='group')
    def test_get(self, base):
        self.assertEqual(Groups.get('groupname'), 'group')

    @mock.patch('pastry.resources.groups.Base.create', return_value='group')
    def test_create(self, base):
        self.assertEqual(Groups.create({}), 'group')

    @mock.patch('pastry.resources.groups.Base.update', return_value='group')
    def test_update(self, base):
        self.assertEqual(Groups.update('group', {}), 'group')

    @mock.patch('pastry.resources.groups.Base.delete', return_value='group')
    def test_delete(self, base):
        self.assertEqual(Groups.delete('group'), 'group')

    @mock.patch('pastry.resources.groups.Base.exists', return_value=True)
    def test_exists(self, base):
        self.assertEqual(Groups.exists('group'), True)



'''Search provides access to chef's search functionality'''

import urllib
from .base import Base
from pastry.pastry_client import PastryClient


class Search(Base):
    '''
    Provides access to chef `search`
    '''
    _base_url = '/organizations/%(org)s/search'

    @classmethod
    def index(cls):
        '''
        Fetches the available search indexes

        :return: The search indexes
        :rtype: hash
        '''
        return super(Search, cls).index()

    @classmethod
    def run(cls, index, query='*:*', rows=1000, start=0, filters=None):
        '''
        Runs the search query against the index. The default will return
        the a list of all of the properties of the match. Use the filters
        to select which fields should be returned by the query.

        The filters should be a hash where the key is the name of the field
        in the returned hash, and the value is a space separated list of
        where in the index to find the value. e.g.::

            {
                'name': [ 'name' ],
                'ip': [ 'ipaddress' ],
                'kernel_version': [ 'kernel', 'version' ]
            }

        :param index: The index to search against
        :param query: The SOLR query to match the index against
        :param rows: The maximum number of rows to return
        :param start: The row to start at
        :param filters: The filters to apply to the result
        :type index: string
        :type query: string
        :type rows: integer
        :type start: integer
        :type filters: hash
        :return: The matching (and filtered) results
        :rtype: hash
        '''
        query_string = cls.escape_query(urllib.urlencode({
            'query': query,
            'rows': rows,
            'start': start
        }))
        return PastryClient.call(
            '%s/%s?%s' % (cls.base_url(), index, query_string),
            method='POST',
            data=filters
        )



'''base is a class that resources inherit from to provide common http
methods'''

from pastry.pastry_client import PastryClient
from pastry.exceptions import HttpError


class Base(object):
    '''
    Base class for chef resources to inherit from
    '''
    _base_url = None

    @classmethod
    def base_url(cls):
        '''
        Get the base url for this resaource - raise a ValueError if the
        child class has not set it's url
        '''
        if not cls._base_url:
            raise ValueError('Base url has not been set of this resource')
        return cls._base_url

    @classmethod
    def index(cls):
        '''
        Fetch all instances of this resource
        '''
        return PastryClient.call(cls.base_url())

    @classmethod
    def get(cls, instanceid):
        '''
        Fetch a specific instance of a resource
        '''
        return PastryClient.call('%s/%s' % (cls.base_url(), instanceid))

    @classmethod
    def create(cls, instance):
        '''
        Create a new instance
        '''
        return PastryClient.call(
            '%s' % cls.base_url(), method='POST', data=instance)

    @classmethod
    def update(cls, instanceid, instance):
        '''
        Update an instance
        '''
        return PastryClient.call(
            '%s/%s' % (cls.base_url(), instanceid),
            method='PUT',
            data=instance
        )

    @classmethod
    def delete(cls, instanceid):
        '''
        Delete an instance
        '''
        return PastryClient.call(
            '%s/%s' % (cls.base_url(), instanceid), method='DELETE')

    @classmethod
    def exists(cls, instanceid):
        '''
        Check if an instance exists
        '''
        try:
            cls.get(instanceid)
        except HttpError as err:
            if err.statuscode == 404:
                return False
        return True

    @classmethod
    def escape_query(cls, query_string):
        '''
        Escape '%' characters in query_string
        '''
        return query_string.replace('%', '%%')

    @classmethod
    def get_acl(cls, instanceid):
        '''
        Fetch the acl for a chef resource
        '''
        return PastryClient.call(
            '%s/%s/_acl' % (cls.base_url(), instanceid))

    @classmethod
    def set_permission(cls, instanceid, permission, actors):
        '''
        Set a permission on a chef resource
        '''
        return PastryClient.call(
            '%s/%s/_acl/%s' % (cls.base_url(), instanceid, permission),
            method='PUT',
            data={permission: actors}
        )
